package com.ibm.oti.shared;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 1998, 2010  All Rights Reserved
 */

import java.net.URL;

/**
 * Implementation of SharedClassURLClasspathHelper.
 * <p>
 * @author		OTI
 * @version		initial
 *
 * @see SharedClassURLClasspathHelper
 * @see SharedClassHelperFactory
 * @see SharedClassAbstractHelper
 */
public class SharedClassURLClasspathHelperImpl extends SharedClassAbstractHelper implements
		SharedClassURLClasspathHelper {
	private URL[] urls, origurls;
	private boolean[] validated;
	private int urlCount, confirmedCount;

	/* Not public - should only be created by factory */
	SharedClassURLClasspathHelperImpl(ClassLoader loader, URL[] classpath, int id, boolean canFind, boolean canStore) {
		this.origurls = classpath;
		this.urls = new URL[classpath.length];
		this.urlCount = classpath.length;
		this.validated = new boolean[classpath.length];
		this.confirmedCount = 0;
		initialize(loader, id, canFind, canStore);
		initializeShareableClassloader(loader);
		initializeURLs();
	}

	private void initializeURLs() {
		for (int i=0; i<urlCount; i++) {
			urls[i] = convertJarURL(origurls[i]);
		}
	}

	private native int findSharedClassImpl2(int id, String partition, String className, ClassLoader loader, URL[] urls,
			boolean doFind, boolean doStore, int urlCount, int confirmedCount, byte[] romClassCookie);

	private native boolean storeSharedClassImpl2(int id, String partition, ClassLoader loader, URL[] urls, int urlCount, int cpLoadIndex, Class clazz, byte[] nativeFlags);

	/* Before setClasspath(), classpath changes were detected by change in urlCount. However, this is now
	 * broken by the addition of setClasspath(). All the classpath cacheing optimizations in the cache
	 * also depend on the urlCount change, which is also broken by setClasspath(). Hence, this native
	 * essentially flushes the classpath caches and forces them to rebuild.
	 */
	private native void notifyClasspathChange(int id);

	public byte[] findSharedClass(String className, IndexHolder indexFoundAtHolder) {
		return findSharedClass(null, className, indexFoundAtHolder);
	}

	public synchronized byte[] findSharedClass(String partition, String className, IndexHolder indexFoundAtHolder) {
		if (!canFind) {
			return null;
		}
		if (className==null) {
			printVerboseError("Cannot call findSharedClass with null class name. Returning null.");
			return null;
		}
		/* Important not to call findSharedClassImpl if confirmedCount==0 as 0 means "confirmedCount not set" */
		if (confirmedCount==0) {
			printVerboseInfo("There are no confirmed elements in the classpath. Returning null.");
			return null;
		}
		SharedClassFilter theFilter = getSharingFilter();
		boolean doFind, doStore;
		if (theFilter!=null) {
			doFind = theFilter.acceptFind(className);
			/* Don't invoke the store filter if the cache is full */
			if (nativeFlags[CACHE_FULL_FLAG] == 0) {
				doStore = theFilter.acceptStore(className);
			} else {
				doStore = true;
			}
		} else {
			doFind = true;
			doStore = true;
		}
		byte[] romClassCookie = new byte[ROMCLASS_COOKIE_SIZE];
		int indexFoundAt = findSharedClassImpl2(this.id, partition, className, getClassLoader(), this.urls, doFind, doStore, this.urlCount, this.confirmedCount, romClassCookie);
		/* indexFoundAt will be -1 if class is not found */
		if (indexFoundAt < 0) {
			return null;
		}
		if (indexFoundAtHolder!=null) {
			indexFoundAtHolder.setIndex(indexFoundAt);
		}
		return romClassCookie;
	}

	public boolean storeSharedClass(Class clazz, int foundAtIndex) {
		return storeSharedClass(null, clazz, foundAtIndex);
	}

	public synchronized boolean storeSharedClass(String partition, Class clazz, int foundAtIndex) {
		if (!canStore) {
			return false;
		}
		if (clazz==null) {
			printVerboseError("Cannot call storeSharedClass with null Class. Returning false.");
			return false;
		}
		if (urlCount==0) {
			printVerboseError("Classpath has zero elements. Cannot call storeSharedClass without classpath. Returning false.");
			return false;
		}
		if (foundAtIndex<0) {
			printVerboseError("foundAtIndex cannot be <0 for storeSharedClass. Returning false.");
			return false;
		}
		if (foundAtIndex>=urlCount) {
			printVerboseError("foundAtIndex cannot be >= urls in classpath for storeSharedClass. Returning false.");
			return false;
		}
		if (!validated[foundAtIndex]) {
			/* Because we only check each element once, we can afford to also check whether the URL exists */
			if (!validateURL(urls[foundAtIndex], true)) {
				return false;
			}
			validated[foundAtIndex]=true;
		}
		ClassLoader actualLoader = getClassLoader();
		if (!validateClassLoader(actualLoader, clazz)) {
			/* Attempt to call storeSharedClass with class defined by a different classloader */
			return false;
		}
		if (confirmedCount <= foundAtIndex) {
			confirmedCount = foundAtIndex+1;
			printVerboseInfo("Number of confirmed entries is now "+confirmedCount);
		}
		return storeSharedClassImpl2(this.id, partition, actualLoader, this.urls, this.urlCount, foundAtIndex, clazz, nativeFlags);
	}

	private synchronized void growURLs(int toMinSize) {
		printVerboseInfo("Growing URL array to "+toMinSize);
		int newSize = ((toMinSize+1)*2);		/* toMinSize could be zero! */
		URL[] newOrigUrls = new URL[newSize];
		URL[] newUrls = new URL[newSize];
		boolean[] newValidated = new boolean[newSize];
		System.arraycopy(origurls, 0, newOrigUrls, 0, urlCount);
		System.arraycopy(urls, 0, newUrls, 0, urlCount);
		System.arraycopy(validated, 0, newValidated, 0, urlCount);
		origurls = newOrigUrls;
		urls = newUrls;
		validated = newValidated;
	}

	public synchronized void addClasspathEntry(URL cpe) {
		if (cpe==null) {
			printVerboseError("URL is null for addClasspathEntry.");
			return;
		}
		/* If array is not big enough, grow it */
		if (urls.length < (urlCount+1)) {
			growURLs(urlCount+1);
		}
		origurls[urlCount] = cpe;
		urls[urlCount] = convertJarURL(cpe);
		++urlCount;

		/* Note: No need to notifyClasspathChange().
		 * Appending a URL changes the urlCount which is picked up by the cache code
		 */
	}

	/* Function required by the factory */
	synchronized URL[] getClasspath() {
		URL[] correctLengthArray = new URL[urlCount];
		System.arraycopy(origurls, 0, correctLengthArray, 0, urlCount);
		return correctLengthArray;
	}

	public synchronized void confirmAllEntries() {
		confirmedCount = urlCount;
	}

	public synchronized void setClasspath(URL[] newClasspath) throws CannotSetClasspathException {
		boolean changeMade = false;

		if (newClasspath.length < confirmedCount) {
			throw new CannotSetClasspathException("New classpath cannot be shorter than confirmed elements of original");
		}
		for (int i=0; i<confirmedCount; i++) {
			if (!newClasspath[i].equals(origurls[i])) {
				throw new CannotSetClasspathException("Index "+i+" of newClasspath does not match confirmed original");
			}
		}
		/* Grow urls and validated if necessary */
		if (newClasspath.length > origurls.length) {
			growURLs(newClasspath.length);
		}
		/* Having ensured that confirmed URLs are the same, copy the others */
		for (int i=confirmedCount; i<newClasspath.length; i++) {
			if (!newClasspath[i].equals(origurls[i])) {
				origurls[i] = newClasspath[i];
				urls[i] = convertJarURL(newClasspath[i]);
				changeMade = true;
			}
			/* validated[i] will already be false as it would be confirmed otherwise, so no need to set */
		}
		/* If new classpath is shorter, remaining entries will be ignored as they are > urlCount */
		if (urlCount != newClasspath.length) {
			urlCount = newClasspath.length;
			changeMade = true;
		}
		/* Only call notifyClasspathChange if a genuine change has been made, otherwise internal data caches
		 * will be cleared unnecessarily */
		if (changeMade) {
			printVerboseInfo("setClasspath() updated classpath. Now urlCount="+urlCount);
			notifyClasspathChange(this.id);
		}
	}

	String getHelperType() {
		return "SharedClassURLClasspathHelper";
	}
}
